@@ -1,35 +1,49 @@
 /*
  *	helper functions for scponly
  */
-#include <stdio.h>	// io
-#include <string.h>	// for str*
-#include <sys/types.h>	// for stat, getpwuid
-#include <sys/stat.h>	// for stat
-#include <unistd.h>	// for exit, access, getpwuid, execve
-#include <errno.h>	// for debugging
-#include <pwd.h>	// to get username for config parsing
-#include <time.h>	// time
-#include <libgen.h>	// basename
-#include <stdlib.h>	// realloc
+#include <stdio.h>	/* io */
+#include <string.h>	/* for str* */
+#include <sys/types.h>	/* for stat, getpwuid */
+#include <sys/stat.h>	/* for stat */
+#include <unistd.h>	/* for exit, access, getpwuid, execve, getopt */
+#ifdef HAVE_GETOPT_H
+#include <getopt.h> /* for getopt */
+#endif
+#include <errno.h>	/* for debugging */
+#include <pwd.h>	/* to get username for config parsing */
+#include <time.h>	/* time */
+#include <libgen.h>	/* basename */
+#include <stdlib.h>	/* realloc */
 #include <syslog.h>
 #include "scponly.h"
 #include "config.h"
 
 #ifdef HAVE_GLOB
-#include <glob.h>	// for glob()
+#include <glob.h>	/* for glob() */
 #else
 #ifdef HAVE_WORDEXP
-#include <wordexp.h>	// for wordexp()
+#include <wordexp.h>	/* for wordexp() */
 #endif
 #endif
 
 #define MAX(x,y)	( ( x > y ) ? x : y )
+#define MIN(x,y)	( ( x < y ) ? x : y )
 
 extern int debuglevel;
 extern char username[MAX_USERNAME];
 extern char homedir[FILENAME_MAX];
 extern cmd_t commands[];
 extern cmd_arg_t dangerous_args[];
+extern char * allowed_env_vars[];
+extern char * safeenv[MAX_ENV];
+
+#ifdef HAVE_GETOPT
+extern char *optarg;
+extern int optind;
+#ifdef HAVE_OPTRESET
+extern int optreset;
+#endif
+#endif
 
 #ifdef UNIX_COMPAT
 char* solaris_needs_strsep(char** str, char* delims)
@@ -109,7 +123,7 @@ char *flatten_vector(char **av)
 
 /*
  *	since some programs support invoking other programs for their encryption
- *	(dropings to replace ssh), we must refuse to support these arguments
+ *	(dropins to replace ssh), we must refuse to support these arguments
  *
  *	RETURN: 1 means reject this command, 0 means it is safe.
  */
@@ -117,6 +131,8 @@ int check_dangerous_args(char **av)
 {
 	cmd_arg_t	*cmdarg=dangerous_args;
 	char		**tmpptr=av;
+	int			ch;
+	int			ac=0;
 
 	while (cmdarg != NULL)
 	{
@@ -128,18 +144,67 @@ int check_dangerous_args(char **av)
 			 *	the command matches one of our dangerous commands
 			 *	check the rest of the vector for dangerous command
 			 *	line arguments
+			 *
+			 *	if the "getoptflag" is set for this command, use getopt
+			 *	to determine if the flag is present, otherwise
+			 *	to a string match on each element of the argument vector
 			 */
-			tmpptr=av;
-			*tmpptr++;
-			while (*tmpptr!=NULL)
+			if (1 == cmdarg->getoptflag)
 			{
-				if(exact_match(*tmpptr, cmdarg->badarg))
-				{
-					syslog(LOG_ERR, "%s is not permitted for use with %s (%s))", 
-						cmdarg->badarg, cmdarg->name, logstamp());
-					return 1;
+				/*	
+				 *	first count the arguments in the vector
+				 */
+				tmpptr=av;
+				while (*tmpptr!=NULL)
+				{	
+					*tmpptr++;
+					ac++;
 				}
+				/* 
+				 *	now use getopt to look for our problem option
+				 */
+#ifdef HAVE_GETOPT
+#ifdef HAVE_OPTRESET
+				/*
+				 * if we have optreset, use it to reset the global variables
+				 */
+				optreset=1;
+				optind=1;
+#else
+				/*
+				 *	otherwise, try a glibc-style reset of the global getopt vars
+				 */
+				optind=0;
+#endif
+				/*
+				 *	tell getopt to only be strict if the 'opts' is well defined
+				 */
+				opterr=cmdarg->strict;
+				while ((ch = getopt(ac, av, cmdarg->opts)) != -1)
+					if (ch == cmdarg->badarg[0])
+					{
+						syslog(LOG_ERR, "option %s is not permitted for use with %s (arg was %s)(%s))", 
+							cmdarg->badarg, cmdarg->name, optarg, logstamp());
+						return 1;
+					}
+#elif
+				syslog(LOG_ERR, "a getopt() argument check could not be performed for %s, recompile scponly without support for %s or rebuild scponly with getopt", av[0], av[0]);
+#endif
+			}
+			else
+			{
+				tmpptr=av;
 				*tmpptr++;
+				while (*tmpptr!=NULL)
+				{
+					if(strbeg(*tmpptr, cmdarg->badarg))
+					{
+						syslog(LOG_ERR, "%s is not permitted for use with %s (%s))", 
+							cmdarg->badarg, cmdarg->name, logstamp());
+						return 1;
+					}
+					*tmpptr++;
+				}
 			}
 		}
 		cmdarg++;
@@ -178,7 +243,7 @@ char *substitute_known_path(char *request)
 			break;
 		if (exact_match(basename(cmd->name),stripped_req))
 		{
-			free(stripped_req); //discard old pathname
+			free(stripped_req); /* discard old pathname */
 			return (strdup(cmd->name));
 		}
 		cmd++;
@@ -196,7 +261,7 @@ char **build_arg_vector(char *request)
 	char **ap2,**av=(char **)malloc(MAX_ARGC * (sizeof(char *)));
 
 	ap=argv;
-	freeme=inputstring=strdup(request); // make a local copy 
+	freeme=inputstring=strdup(request); /* make a local copy  */
 
         while (ap < &argv[(MAX_ARGC-1)]) 
 	{
@@ -251,7 +316,7 @@ char **expand_wildcards(char **av_old)
 {
 	char		**av_new=(char **)malloc(MAX_ARGC * (sizeof(char *)));
 	glob_t g;
-	int c_old,c_new,c;	// argument counters
+	int c_old,c_new,c;	/* argument counters */
 #ifdef UNIX_COMPAT
 	int flags = GLOB_NOCHECK;
 #else
@@ -394,3 +459,46 @@ int get_uservar(void)
 	return 1;
 }
 
+int mysetenv(const char *name, const char *value) {
+	/* from: http://www.onlamp.com/pub/a/onlamp/excerpt/PUIS3_chap16/index3.html */
+	static char count = 0;
+	char buff[255];
+	
+	if (count == 0)
+		/* in case nothing ever gets put in here... */
+		safeenv[0] = NULL;
+	if (count == MAX_ENV)
+			return 0;
+	if (!name || !value)
+			return 0;
+	if (snprintf(buff, sizeof(buff), "%s=%s", name, value) < 0)
+			return 0;
+	if (safeenv[count] = strdup(buff)) {
+		safeenv[++count] = NULL;
+		return 1;
+	}
+	return 0;
+}
+
+void filter_allowed_env_vars() {
+	
+	int slen = 0;
+	char *p_env, *p_str;
+	char **p_valid = allowed_env_vars;
+	
+	/* check each allowed variable */
+	while (NULL != *p_valid) {
+		
+		p_env = (char*)getenv(*p_valid);
+		if (NULL != p_env) {
+			if (debuglevel)
+				syslog(LOG_DEBUG, "Found \"%s\" and setting it to \"%s\"", *p_valid, p_env);
+			if (!mysetenv(*p_valid, p_env))
+				syslog(LOG_ERR, "Unable to set environment var \"%s\" to \"%s\"", *p_valid, p_env);
+		} else if (debuglevel) {
+			syslog(LOG_DEBUG, "Unable to find \"%s\" in the environment", *p_valid);
+		}
+		p_valid++;
+	}
+}
+
