@@ -7,12 +7,13 @@
  *	see CONTRIB for additional credits
  */
  
-#include <stdio.h>	// io
-#include <string.h>	// for str*
-#include <sys/types.h>	// for fork, wait
-#include <sys/wait.h>	// for wait
-#include <unistd.h>	// for exit, fork
-#include <stdlib.h>	// EXIT_*
+#include <stdio.h>	/* io */
+#include <string.h>	/* for str* */
+#include <sys/types.h>	/* for fork, wait, stat */
+#include <sys/stat.h> /* for stat */
+#include <sys/wait.h>	/* for wait */
+#include <unistd.h>	/* for exit, fork */
+#include <stdlib.h>	/* EXIT_* */
 #include <errno.h>
 #include <syslog.h>
 #include "scponly.h"
@@ -23,13 +24,13 @@ int chrooted=0;
 char username[MAX_USERNAME];
 char homedir[FILENAME_MAX];
 char chrootdir[FILENAME_MAX];
+char *safeenv[MAX_ENV];
 
 cmd_t commands[] =
 { 
 #ifdef ENABLE_SFTP
 	{ PROG_SFTP_SERVER, 0 },
 #endif /*ENABLE_SFTP*/
-
 #ifdef ENABLE_SCP2
 	{ PROG_LS, 1 }, 
 	{ PROG_CHMOD, 1 },
@@ -62,10 +63,18 @@ cmd_t commands[] =
 	{ PROG_PASSWD, 1 },
 #endif /*ENABLE_PASSWD*/
 
+#ifdef QUOTA_COMPAT
+	{ PROG_QUOTA, 1 },
+#endif /*QUOTA_COMPAT*/
+
 #ifdef SVN_COMPAT
 	{ PROG_SVN, 1 },
 #endif /*ENABLE_SVN*/
 
+#ifdef SVNSERV_COMPAT
+	{ PROG_SVNSERV, 1 },
+#endif /*ENABLE_SVNSERV*/
+	
 	NULL
 };
 
@@ -76,16 +85,53 @@ cmd_t commands[] =
  */
 cmd_arg_t dangerous_args[] =
 {
-	{ PROG_SCP, "-S" },
-	{ PROG_SFTP_SERVER, "-S" },
-#ifdef UNISON_COMPAT
-	{ PROG_UNISON, "-rshcmd" },
-	{ PROG_UNISON, "-sshcmd" },
-	{ PROG_UNISON, "-servercmd" },
+	/*
+	 *	'oplist' only neccesary where 'use getopt' is 1
+	 *	'strict optlist' only applicable where 'use getopt?' is 1
+	 *
+	 * program name		use getopt?		strict optlist?	optname			optlist
+	 */
+#ifdef ENABLE_SCP2
+	{ PROG_SCP, 		1, 				1,				"S",			"dfl:prtvBCc:i:P:q1246S:o:F:" },
 #endif
 #ifdef RSYNC_COMPAT
-	{ PROG_RSYNC, "-e" },
-#endif 
+	{ PROG_RSYNC, 		1, 				0,				"e",			"e:" },
+	{ PROG_RSYNC, 		0, 				0,				"-rsh",			NULL },
+#endif 	
+#ifdef UNISON_COMPAT	
+	{ PROG_UNISON, 		0, 				0,				"-rshcmd",		NULL },
+	{ PROG_UNISON, 		0, 				0,				"-sshcmd",		NULL },
+	{ PROG_UNISON, 		0, 				0,				"-servercmd,	NULL" },
+#endif
+	NULL
+};
+
+/*
+ *	SFTP logging requires that the following environment variables be
+ *	defined in order to work:
+ *
+ *	LOG_SFTP
+ *	USER
+ *	SFTP_UMASK
+ *	SFTP_PERMIT_CHMOD
+ *	SFTP_PERMIT_CHOWN
+ *	SFTP_LOG_LEVEL
+ *	SFTP_LOG_FACILITY
+ */
+char * allowed_env_vars[] =
+{
+#ifdef SFTP_LOGGING
+	"LOG_SFTP",
+	"USER",
+	"SFTP_UMASK",
+	"SFTP_PERMIT_CHMOD",
+	"SFTP_PERMIT_CHOWN",
+	"SFTP_LOG_LEVEL",
+	"SFTP_LOG_FACILITY",
+#endif
+#ifdef UNISON_COMPAT
+	"HOME",
+#endif
 	NULL
 };
 
@@ -98,7 +144,8 @@ int main (int argc, char **argv)
 {
 	FILE *debugfile;
 	int logopts = LOG_PID|LOG_NDELAY;
-	
+	struct stat	homedirstat;
+
 	/*
 	 * set debuglevel.  any nonzero number will result in debugging info to log
 	 */
@@ -108,7 +155,7 @@ int main (int argc, char **argv)
 		fclose(debugfile);
 	}
 #ifndef UNIX_COMPAT
-	if (debuglevel > 1) // debuglevel 1 will still log to syslog
+	if (debuglevel > 1) /* debuglevel 1 will still log to syslog */
 		logopts |= LOG_PERROR;
 #endif
 
@@ -117,7 +164,7 @@ int main (int argc, char **argv)
 #elif IRIX_COMPAT
         openlog(PACKAGE_NAME, logopts, LOG_AUTH);
 #else
-        if (debuglevel > 1) // debuglevel 1 will still log to syslog
+        if (debuglevel > 1) /* debuglevel 1 will still log to syslog */
                 logopts |= LOG_PERROR;
         openlog(PACKAGE_NAME, logopts, LOG_AUTHPRIV);
 #endif
@@ -138,7 +185,7 @@ int main (int argc, char **argv)
 			syslog(LOG_INFO, "chrooted binary in place, will chroot()");
 		chrooted=1;
 	}
-#endif //CHROOTED_NAME
+#endif /* CHROOTED_NAME */
 
 	if (debuglevel)
 	{
@@ -180,17 +227,48 @@ int main (int argc, char **argv)
 	if (chrooted)
 	{
 		char *root_dir = chrootdir;
+		char chdir_path[FILENAME_MAX];
 
 		strcpy(chrootdir, homedir);
+		strcpy(chdir_path, "/");
 		while((root_dir = strchr(root_dir, '/')) != NULL) 
 		{
 			if (strncmp(root_dir, "//", 2) == 0) 
 			{
+				snprintf(chdir_path, FILENAME_MAX, "%s", root_dir + 1);
 				*root_dir = '\0';
 				break;
 			}
 			root_dir++;
 		}
+#ifdef CHROOT_CHECKDIR
+		bzero(&homedirstat, sizeof(struct stat));
+		if (-1 == stat(chrootdir, &homedirstat))
+		{
+			syslog (LOG_ERR, "couldnt stat chroot dir: %s with errno %u", chrootdir, errno);
+			exit(EXIT_FAILURE);
+		}
+		if (0 == (homedirstat.st_mode | S_IFDIR))
+		{
+			syslog (LOG_ERR, "chroot dir is not a directory: %s", chrootdir);
+			exit(EXIT_FAILURE);
+		}
+		if (homedirstat.st_uid != 0)
+		{
+			syslog (LOG_ERR, "chroot dir not owned by root: %s", chrootdir);
+			exit(EXIT_FAILURE);
+		}
+		if (0 != (homedirstat.st_mode & S_IWOTH))
+		{
+			syslog (LOG_ERR, "chroot dir writable by other: %s", chrootdir);
+			exit(EXIT_FAILURE);
+		}
+		if (0 != (homedirstat.st_mode & S_IWGRP))
+		{
+			syslog (LOG_ERR, "chroot dir writable by group: %s", chrootdir);
+			exit(EXIT_FAILURE);
+		}
+#endif
 		if (debuglevel)
 			syslog (LOG_DEBUG, "chrooting to dir: \"%s\"", chrootdir);
 		if (-1==(chroot(chrootdir)))
@@ -202,8 +280,21 @@ int main (int argc, char **argv)
 			syslog (LOG_ERR, "couldn't chroot to %s [%s]", chrootdir, logstamp());
 			exit(EXIT_FAILURE);
 		}
+	    if (debuglevel)                                                                                                
+		{                                                                                                              
+			syslog (LOG_DEBUG, "chdiring to dir: \"%s\"", chdir_path);                                             
+		}                                                                                                              
+		if (-1==(chdir(chdir_path)))                                                                                   
+		{                                                                                                              
+			if (debuglevel)                                                                                        
+			{                                                                                                      
+				syslog (LOG_ERR, "chdir: %m");                                                                 
+			}                                                                                                      
+			syslog (LOG_ERR, "couldn't chdir to %s [%s]", chdir, logstamp());                                      
+			exit(EXIT_FAILURE);     
+		}
 	}
-#endif //CHROOTED_NAME
+#endif /* CHROOTED_NAME */
 
 	if (debuglevel)
 		syslog (LOG_DEBUG, "setting uid to %u", getuid());
@@ -225,7 +316,7 @@ int main (int argc, char **argv)
 		}
 	}
 #else
-	if (0)	{}	// placeholder
+	if (0)	{}	/*  placeholder */
 #endif
 	else if (-1==process_ssh_request(argv[2]))
 	{
@@ -253,9 +344,9 @@ int winscp_transit_request(char *request)
 	new_request=strbeg(request,WINSCP_BOF_REQ);
 	if (NULL == new_request)
 	{
-		return(-1);	 // improper transit cmd
+		return(-1);	 /* improper transit cmd */
 	}
-	printf ("%s\n",WINSCP_BOF); // start transfer
+	printf ("%s\n",WINSCP_BOF); /* start transfer */
 	fflush(stdout);
 	return(winscp_regular_request(new_request));
 }
@@ -281,7 +372,7 @@ int winscp_regular_request(char *request)
 			{
 				printf ("command wasn't terminated with %s, %s or %s\n",
 					WINSCP_EOF_REQ_RETVAL, WINSCP_EOF_REQ_ZERO, WINSCP_EOF_REQ_STATUS);
-				return(-1);	// bogus termination
+				return(-1);	/* bogus termination */
 			}
 		}
 	}
@@ -303,7 +394,7 @@ int winscp_regular_request(char *request)
 	else
 	{
 		retval=process_ssh_request(new_request);
-		if (retzero)		// ignore actual retval if winscp wants us to
+		if (retzero)		/* ignore actual retval if winscp wants us to */
 			retval=0;
 	}
 	free(new_request);
@@ -313,7 +404,7 @@ int winscp_regular_request(char *request)
 int process_winscp_requests(void)
 {
         char    linebuf[MAX_REQUEST];
-	int	count=0;		// num of semicolons in cmd
+	int	count=0;		/* num of semicolons in cmd */
 	int	ack=0;
 
 	winscp_mode=1;
@@ -335,21 +426,21 @@ int process_winscp_requests(void)
 		if (strlen(linebuf)==0)
 			return(-1);
 
-		linebuf[strlen(linebuf)-1]=0;	// drop the trailing CR
+		linebuf[strlen(linebuf)-1]=0;	/* drop the trailing CR */
 		count=cntchr(linebuf,';');
 
-		if (count==1)	// regular cmd
+		if (count==1)	/* regular cmd */
 		{
 			ack=winscp_regular_request(linebuf);
 		}
-		else if (count==2)	 // transit command
+		else if (count==2)	 /* transit command */
 		{
 			ack=winscp_transit_request(linebuf);
 		}
 		else
-			ack=0; 	// winscp always sends 2 or 3 cmds at once
+			ack=0; 	/* winscp always sends 2 or 3 cmds at once */
 
-		printf ("%s%d\n",WINSCP_EOF,ack); // respond to client
+		printf ("%s%d\n",WINSCP_EOF,ack); /* respond to client */
 		fflush(stdout);
 	}
 	return 0;
@@ -359,7 +450,7 @@ int process_winscp_requests(void)
 
 int process_ssh_request(char *request)
 {
-	char **av;
+	char **av, **tenv;
 	char *flat_request,*tmpstring, *tmprequest;
 	char bad_winscp3str[] = "test -x /usr/lib/sftp-server && exec /usr/lib/sftp-server test -x /usr/local/lib/sftp-server && exec /usr/local/lib/sftp-server exec sftp-server";
 	int retval;
@@ -488,30 +579,64 @@ int process_ssh_request(char *request)
 	if (valid_arg_vector(av))
 	{
 
-#ifdef UNISON_COMPAT
-		if (((strlen(homedir) + 6 ) > FILENAME_MAX) ||
-			(-1 == asprintf( &env[0], "HOME=%s", homedir)))
-		{
-			syslog(LOG_ERR, "could not set HOME environment variable(%s))", logstamp());
-			exit(EXIT_FAILURE);
+/*                                                                                                                   
+ * Unison needs the HOME environment variable be set to the directory                                                  
+ * where the .unison directory resides.                                                                                
+ */                                                                                                                    
+#ifdef USE_SAFE_ENVIRONMENT
+		safeenv[0] = NULL;
+		filter_allowed_env_vars();
+		tenv = safeenv;
+		if (debuglevel) {
+			while (NULL != *tenv) {
+				syslog(LOG_DEBUG, "Environment contains \"%s\"", *tenv++);
+			}
 		}
-		if (debuglevel)
-			syslog(LOG_DEBUG, "set HOME environment variable to %s (%s))", env[0], logstamp());
-#endif 
+#endif
 
+#ifdef UNISON_COMPAT
+		if (chrooted)                                                                                           
+		{                                                                                                       
+			if (!mysetenv("HOME","/"))                                                                       
+			{                                                                                               
+				syslog(LOG_ERR, "could not set HOME environment variable to '/' (%s)", logstamp());           
+				exit(EXIT_FAILURE);                                                                     
+			}                                                                                               
+			if (debuglevel)                                                                                 
+				syslog(LOG_DEBUG, "set HOME environment variable to / %s", logstamp());           
+		}                                                                                                       
+		else
+		{
+			if (((strlen(homedir) + 6 ) > FILENAME_MAX) || !mysetenv("HOME",homedir))
+			{
+				syslog(LOG_ERR, "could not set HOME environment variable (%s)", logstamp());
+				exit(EXIT_FAILURE);
+			}
+			if (debuglevel)
+				syslog(LOG_DEBUG, "set HOME environment variable to %s (%s)", homedir, logstamp());
+		}
+#endif 
 		syslog(LOG_INFO, "running: %s (%s)", flat_request, logstamp());
+
 #ifdef WINSCP_COMPAT
 		if (winscp_mode)
 		{
 			int status=0;
 			if (fork() == 0)
-				retval=execve(av[0],av,env);
+#ifdef USE_SAFE_ENVIRONMENT
+				retval=execve(av[0],av,safeenv);
+#else
+				retval=execve(av[0],av,NULL);
+#endif
 			else
 			{
 				wait(&status);
 				fflush(stdout);
 				fflush(stderr);
 				discard_vector(av);
+#ifdef USE_SAFE_ENVIRONMENT                                                                                            
+				discard_vector(safeenv);
+#endif
 				free(flat_request);
 				free(tmprequest);
 				return(WEXITSTATUS(status));
@@ -520,11 +645,18 @@ int process_ssh_request(char *request)
 		else
 #endif
 		{
+#ifdef USE_SAFE_ENVIRONMENT
+			retval=execve(av[0],av,safeenv);
+#else
 			retval=execve(av[0],av,NULL);
+#endif
 		}
 		syslog(LOG_ERR, "failed: %s with error %s(%u) (%s)", flat_request, strerror(errno), errno, logstamp());
 		free(flat_request);
 		discard_vector(av);
+#ifdef USE_SAFE_ENVIRONMENT                                                                                            
+		discard_vector(safeenv);
+#endif
 #ifdef WINSCP_COMPAT
 		if (winscp_mode)
 		{
