@@ -1,12 +1,13 @@
-#include <stdio.h>	// FILENAME_MAX
+#include <stdio.h>	/* FILENAME_MAX */
 #include "config.h"
 
 #define MAX_USERNAME 32
-#define MAX_REQUEST (1024)		// any request exceeding this is truncated
-#define MAX_ARGC 100			// be reasonable
+#define MAX_REQUEST (1024)		/* any request exceeding this is truncated */
+#define MAX_ARGC 100			/* be reasonable */
+#define MAX_ENV 50				/* be reasonable */
 #define ALLOWABLE "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890 ./-*\\_\'\":?()"
-#define WHITE " \t"			// whitespace
-#define exact_match(x,y) (0==strcmp(x,y))	// we dont want strncpy for this
+#define WHITE " \t"			/* whitespace */
+#define exact_match(x,y) (0==strcmp(x,y))	/* we dont want strncpy for this */
 #define LOGIDENT "scponly"
 
 #define WINSCP_EOF_REQ_RETVAL	" ; echo \"WinSCP: this is end-of-file:$?\""
@@ -26,7 +27,10 @@ typedef struct
 typedef struct
 {
 	char *name;
+	int  getoptflag;
+	int  strict;
 	char *badarg;
+	char *opts;
 } cmd_arg_t;
  
 /*
